// src/backend/concrete_simulation.cc - Copyright 2005, 2006, University
//               of Padova, dept. of Pure and Applied
//               Mathematics
//
// This file is part of SGPEMv2.
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// SGPEMv2 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SGPEMv2; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

#include "concrete_simulation.hh"
#include "simulation_observer.hh"
#include "scheduler.hh"
#include "cpu_policies_gatekeeper.hh"
#include <glibmm/timer.h>

#include <cassert>

#include "smartp.tcc"

#include <algorithm>
#include <cassert>
#include <functional>
#include <iostream>

using namespace std;
using namespace sgpem;
using namespace memory;
using Glib::usleep;

ConcreteSimulation::ConcreteSimulation() :
    _state(state_stopped), _mode(true), _timer_interval(1000), _policy(NULL)
{
	_notify = false;
	_front = 0;
}

void
ConcreteSimulation::set_timer(unsigned int t)
{
  _timer_interval = t;
}

int
ConcreteSimulation::get_timer() const
{
  return _timer_interval;
}

void
ConcreteSimulation::set_mode(const bool& b)
{
  _mode = b;
}

bool
ConcreteSimulation::get_mode() const
{
  return _mode;
}

void
ConcreteSimulation::jump_to(History::position p)
{

  switch (_state)
  {
  case state_running:
    pause();
		break;
  case state_stopped:
    _history.reset(true);
		_front = 0;
    break;
  default:
    ;
  }

	pause();

	bool yet_to_finish = true;
	while (yet_to_finish && p > _front)
		yet_to_finish = step();

	if (!yet_to_finish)
		stop();
	_front = p < _front ? p : _front;
	notify_change();
}


void
ConcreteSimulation::pause()
{
  _state = state_paused;
}

void
ConcreteSimulation::stop()
{
  _state = state_stopped;
}

void
ConcreteSimulation::run() throw(UserInterruptException, NullPolicyException, MalformedPolicyException)
{
  switch (_state)
  {
  case state_running:
    return;
  case state_stopped:
    _history.reset(true);
		_front = 0;
    break;
  default:
    ;
  }

  _state = state_running;

  if (get_policy() == NULL)
  {
    stop();
    throw NullPolicyException("no policy selected");
  }

  //*******  CONTINUOUS TIME
  if (_mode)
  {
    do
    {
      try
      {
        //step forward
	      bool yet_to_finish = true;
				if (_front == get_history().get_size() - 1)
					yet_to_finish = Scheduler::get_instance().step_forward(_history, *get_policy());
        if (!yet_to_finish) stop();
				_front++;
			  notify_change();

        //sleep
        Glib::usleep(_timer_interval*1000);
      }
      catch (const CPUPolicyException& e)
      {
        stop();
        throw;
      }

      //check the state
      if (_state == state_stopped || _state == state_paused)
        return;
    }
    while (true);
  }

  //******* STEP by STEP
  else
  {
    try
    {
      assert(get_policy() != NULL);
      //step forward
      bool yet_to_finish = true;
			if (_front == get_history().get_size() - 1)
				yet_to_finish = Scheduler::get_instance().step_forward(_history, *get_policy());
			_front++;
			notify_change();
      if (yet_to_finish)
        pause();
      else
        stop();
    }
    catch (const CPUPolicyException& e)
    {
      stop();
      throw;
    }
  }
}

bool
ConcreteSimulation::step()
	throw(UserInterruptException, NullPolicyException, MalformedPolicyException)
{
  if (get_policy() == NULL)
  {
    stop();
    throw NullPolicyException("no policy selected");
  }

    try
    {
      assert(get_policy() != NULL);
      //step forward
      bool yet_to_finish = true;
			if (_front == get_history().get_size() - 1)
				yet_to_finish = Scheduler::get_instance().step_forward(_history, *get_policy());
			_front++;
			return yet_to_finish;
    }
    catch (const CPUPolicyException& e)
    {
      stop();
      throw;
    }
}


Simulation::state
ConcreteSimulation::get_state() const
{
  return _state;
}

ConcreteHistory&
ConcreteSimulation::get_history()
{
  return _history;
}

const ConcreteHistory&
ConcreteSimulation::get_history() const
{
  return _history;
}


void
ConcreteSimulation::set_policy(CPUPolicy* p) throw(CPUPolicyException)
{
  // NOTE: restoring of the previous policy is done here because I
  // couldn't think of a clean way to do it
  // inside activate_policy()
  
  try
  {
    CPUPoliciesGatekeeper::get_instance().activate_policy(&_history, p);
    _policy = p;
  }
  catch(const CPUPolicyException& e1)
  {
    try
    {
      // this is a no-op if _policy is NULL
      CPUPoliciesGatekeeper::get_instance().activate_policy(&_history, _policy);
    }
    catch(const CPUPolicyException& e2)
    {
      _policy = NULL;

      string msg = _("unable to change policy and to restore the previous: ");
      msg += e2.what();

      throw CPUPolicyException(msg);
    }

    string msg = _("unable to change policy: ");
    msg+= e1.what();
    
    throw CPUPolicyException(msg);
  }
}


CPUPolicy*
ConcreteSimulation::get_policy()
{
  return _policy;
}
